"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#Signal
         instanceVariableNames:'mayProceed notifierString nameClass message
                                handlerBlock'
         classVariableNames:'NoHandlerSignal'
         poolDictionaries:''
         category:'Kernel-Exceptions'
!

Signal comment:'
COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved
'!

!Signal class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /cvs/stx/stx/libbasic/Signal.st,v 1.10 1994-06-02 16:22:04 claus Exp $
"
!

documentation
"
    Signal and Exception provide a framework for exception handling.

    A Signal object is usually defined somewhere up in the calling chain
    and associated with some abnormal event. Many signals are also
    created at startup time and reused.

    When the event is raised (by Signal>>raise) the control will be either 
    given to a debugger or - if a handler was defined - to the handler. 
    The handler will get a description of what (and where) happened in an 
    Exception object and decide how to react on the situation (i.e. proceed,
    return or restart).

    There is also a companion class called SignalSet, which allows handling
    multiple signals with one handler (for example all arithmetic signals).
    And, finally there is a very special SignalSet which allows catching
    any signal (SignalSet>>anySignal).

    This Signal implementation has been modeled after what some PD
    programs seem to expect - it may not be perfect currently
    (especially, I dont know what nameClass and message are for).

    In addition to the nested catch & throw mechanism, signals can also be
    used when no such scoping exists. To support this, signals can be assigned
    a handlerBlock, which gets evaluated with the exception argument and the
    raise-context, in case no handler was found.
    If neither handler-context, nor handler block is defined, the emergencyHandler
    (from Exception) is evaluated - this one will bring up a debugger by default.

    HandlerBlock allows to globally catch UNIX signals at any time.

    Part of the implementation is a left-over from old times when the resume/
    restart things in context did not work properly. Now, with the handler-
    and suspendedContext at hand, the exception can do it using other mechanisms.
    This might be cleaned up ...

    See samples in doc/coding.

    Instance variables:

        mayProceed      <Boolean>       hint for the debugger - program may proceed
                                        (currently not honored by the debugger)

        notifierString  <String>        eror message to be output 

        nameClass       <???>           I dont know what this is for 
                                        (included for ST-80 compatibility)

        message         <???>           I dont know what this is for
                                        (included for ST-80 compatibility)

        handlerBlock    <Block>         if nonNil, a (2-arg) block to be evaluated
                                        when no handler context is found.
"
! !

!Signal class methodsFor:'initialization'!

initialize
    "setup the signal used to handle unhandled signals"

    NoHandlerSignal isNil ifTrue:[
        NoHandlerSignal := (Signal new) mayProceed:true.
        NoHandlerSignal notifierString:'no Handler'
    ]
! !

!Signal class methodsFor:'instance creation'!

new
    "return a new signal"

    ^ self basicNew notifierString:'signal'
! !

!Signal class methodsFor:'signal access'!

noHandlerSignal
    "return the signal used to handle unhandled signals"

    ^ NoHandlerSignal
! !

!Signal methodsFor:'instance creation'!

newSignalMayProceed:aBoolean
    "create a new signal, using the receiver as a prototype"

    ^ (self copy) mayProceed:aBoolean
!

newSignal
    "create a new signal, using the receiver as a prototype"

    ^ (self copy)
! !

!Signal methodsFor:'accessing'!

nameClass:aClass message:aSelector
    "I dont know what that is used for (yet)"

    nameClass := aClass.
    message := aSelector
!

mayProceed:aBoolean
    "set/clear the signals ability to proceed.
     This flag is not checked in the current version of
     the debugger."

    mayProceed := aBoolean
!

notifierString:aString
    "set the notifier string"

    notifierString := aString
!

notifierString
    "return the notifier string"

    ^ notifierString
!

handlerBlock:aBlock
    "set the handlerblock"

    handlerBlock := aBlock
! !

!Signal methodsFor:'save evaluation'!

handle:handleBlock do:aBlock
    "evaluate the argument, aBlock.
     If the receiver-signal is raised during evaluation,
     evaluate the handleBlock passing it an Exception argument.
     The handler may decide how to react to the signal by sending
     a corresponding message to the exception (see there).
     If the signal is not raised, return the value of evaluating
     aBlock."

     ^ aBlock value  "the real logic is in raise/Exception"
!

catch:aBlock
     "evaluate the argument, aBlock; return false.
      If the receiver-signal is raised during evaluation, abort
      the evaluation and return true. This is the catch & throw
      mechanism found in other languages."

      |raiseOccurred|

      raiseOccurred := false.
      self handle:[:ex | raiseOccurred := true. ex return] do:aBlock.
      ^ raiseOccurred

      "
       Object messageNotUnderstoodSignal catch:[
          123 asciiValue   
       ]
      "
! !

!Signal methodsFor:'queries'!

isHandled
    "return true, if there is a handler for the receiver signal.
     Raising an unhandled signal will usually lead into the debugger,
     but can be cought globally by setting Exceptions EmergencyHandler."

    |con|

    con := thisContext.
    con := con sender.
    [con notNil] whileTrue:[
        (con selector == #handle:do:) ifTrue:[
            "
             is this is the Signal>>handle:do: context
             or a SignalSet>>handle:do: context with self in it ?
            "
            ((con receiver == self) 
            or:[(con receiver isMemberOf:SignalSet) and:[con receiver includes:self]]) ifTrue:[
                "found it"

                ^ true
            ].
        ].
        con := con sender
    ].
    ^ false
! !

!Signal methodsFor:'raising'!

raise
    "raise a signal - create an Exception object
     and call the handler with this as argument.
     The signals notifierString is used as errorString."

    "This could have been defined using 'raiseRequestWith:', 
     but is not - to not add too many contexts to the backtrace 
     (thus making things cleaner in the debugger-walkback eventually)"

    |ex block|

    ex := Exception new 
              signal:self
              parameter:nil 
              errorString:notifierString
              suspendedContext:thisContext sender.

    ex resumeBlock:[:value | ^ value].
    self evaluateHandlerWith:ex.

    "
     finally,
     take either handlerBlock or emergencyHandler ...
    "
    (block := handlerBlock) isNil ifTrue:[
        block := Exception emergencyHandler
    ].

    "... and call it"
    ^ block value:ex value:thisContext.
!

raiseRequestWith:aParameter
    "raise a signal - create an Exception object with aParameter
     and call the handler with this as argument..
     The signals notifierString is used as errorString."

    "This could have been defined using 'raiseRequestWith:', 
     but is not - to not add too many contexts to the backtrace 
     (thus making things cleaner in the debugger-walkback eventually)"

    |ex block|

    ex := Exception new 
              signal:self
              parameter:aParameter 
              errorString:notifierString
              suspendedContext:thisContext sender.

    ex resumeBlock:[:value | ^ value].
    self evaluateHandlerWith:ex.

    "
     finally,
     take either handlerBlock or emergencyHandler ...
    "
    (block := handlerBlock) isNil ifTrue:[
        block := Exception emergencyHandler
    ].

    "... and call it"
    ^ block value:ex value:thisContext.
!

raiseFrom:something
    "raise a signal - create an Exception object with aParameter
     and call the handler with this as argument..
     The printString of the argument, something is used as errorString.
     XXX: I am not certain, if this is the correct behavior (seen in remoteInvocation-goodie)"

    |ex block|

    ex := Exception new 
              signal:self
              parameter:nil
              errorString:something printString
              suspendedContext:thisContext sender.

    ex resumeBlock:[:value | ^ value].
    self evaluateHandlerWith:ex.

    "
     finally,
     take either handlerBlock or emergencyHandler ...
    "
    (block := handlerBlock) isNil ifTrue:[
        block := Exception emergencyHandler
    ].

    "... and call it"
    ^ block value:ex value:thisContext.
!

raiseRequestWith:aParameter errorString:aString
    "raise a signal - create an Exception object with aParameter
     and call the handler with this as argument..
     The argument, aString is used as errorString."

    |ex block|

    ex := Exception new 
              signal:self
              parameter:aParameter 
              errorString:aString
              suspendedContext:thisContext sender.

    ex resumeBlock:[:value | ^ value].
    self evaluateHandlerWith:ex.

    "
     finally,
     take either handlerBlock or emergencyHandler ...
    "
    (block := handlerBlock) isNil ifTrue:[
        block := Exception emergencyHandler
    ].

    "... and call it"
    ^ block value:ex value:thisContext.
! !

!Signal methodsFor:'private'!

evaluateHandlerWith:anException
    "search through the context-calling chain for a 'handle:do:'-context 
     to the receiver or a SignalSet which includes the receiver.
     If found, take its 2nd argument (the handler) and evaluate
     it with the exception as argument.
     If none found, just return."

    |con|

    con := thisContext.
    con := con sender.
    con isRecursive ifTrue:[
        ^ self enterDebuggerWith:'recursive signal raise'
    ].

    [con notNil] whileTrue:[
        (con selector == #'handle:do:') ifTrue:[
            "
             if this is the Signal>>handle:do: context
             or a SignalSet>>handle:do: context with self in it,
             call the handler
            "
            ((con receiver == self) 
            or:[(con receiver isMemberOf:SignalSet) 
                and:[con receiver includes:self]]) ifTrue:[
                "call the handler"

                anException handlerContext:con.
                self doCallHandler:(con args at:1) with:anException.

                "if the handler rejects or falls through we arrive here"
                "continue search for another handler"
            ].
        ].
        con := con sender
    ].

    "
     we arrive here, if either no handler was found, or none of the
     handlers did a return (i.e. every handler rejected or fell through).
    "
!

doCallHandler:aHandler with:ex
    "call the handler proper - needed an extra method
     to have a separate returnContext for the rejectBlock.
     (which is historical, and actually no longer needed)"

    ex rejectBlock:[^ self]. "this will return on reject"
    aHandler value:ex.
    "handler return - is just like a reject"
! !
